use crate::traits::XVar;

/// Trait that should be implemented for objects that will be minimzed
pub trait ObjectiveFun<T: XVar<E> + Clone, E> {
    /// Return the objective function value at a specified coordinate
    fn eval(&self, xs: &T) -> f64;
    /// Return the gradients of the objective function value for specified coordinates
    fn prime(&self, xs: &T) -> E;
}

/// Enum allowing for selection of style of numerical differentiation
pub enum DiffStyle {
    ForwardDifference,
    BackwardDifference,
    CentralDifference,
}

/// Struct that wraps a lambda and provides a numerical derivative for it for use in gradient
/// descent algorithms
pub struct FunWithNumericalDiff {
    function: Box<dyn Fn(&[f64]) -> f64>,
    dx: f64,
    style: DiffStyle,
}

impl ObjectiveFun<Vec<f64>, Vec<f64>> for FunWithNumericalDiff {
    fn eval(&self, xs: &Vec<f64>) -> f64 {
        (self.function)(xs)
    }

    fn prime(&self, xs: &Vec<f64>) -> Vec<f64> {
        let mut xs_local = Vec::new();
        xs_local.extend_from_slice(xs);
        let f: Box<dyn FnMut((usize, &f64)) -> f64> = match self.style {
            DiffStyle::ForwardDifference => Box::new(move |(i, x)| -> f64 {
                xs_local[i] = x + self.dx;
                let xprime = ((self.function)(&xs_local) - (self.function)(xs)) / (self.dx);
                xs_local[i] = *x;
                xprime
            }),
            DiffStyle::BackwardDifference => Box::new(move |(i, x)| -> f64 {
                xs_local[i] = x - self.dx;
                let xprime = ((self.function)(xs) - (self.function)(&xs_local)) / (self.dx);
                xs_local[i] = *x;
                xprime
            }),
            DiffStyle::CentralDifference => Box::new(move |(i, x)| -> f64 {
                xs_local[i] = x - (0.5 * self.dx);
                let f1 = (self.function)(&xs_local);
                xs_local[i] = x + (0.5 * self.dx);
                let f2 = (self.function)(&xs_local);
                xs_local[i] = *x;
                (f2 - f1) / self.dx
            }),
        };
        xs.iter().enumerate().map(f).collect()
    }
}

/// Struct that wraps two lambda with one providing the objective function evaluation and the other
/// providing the gradient value
pub struct Fun<T: XVar<E>, E> {
    function: Box<dyn Fn(&T) -> f64>,
    prime: Box<dyn Fn(&T) -> E>,
}

// Simple type to remove the generics
pub type F64Fun = Fun<f64, f64>;

impl<T: XVar<E>, E> ObjectiveFun<T, E> for Fun<T, E> {
    fn eval(&self, xs: &T) -> f64 {
        (self.function)(xs)
    }

    fn prime(&self, xs: &T) -> E {
        (self.prime)(xs)
    }
}

impl<T: XVar<E>, E> Fun<T, E> {
    pub fn new(function: Box<dyn Fn(&T) -> f64>, prime: Box<dyn Fn(&T) -> E>) -> Self {
        Fun { function, prime }
    }
}
